unit unitResFile;

interface

uses Windows, Classes, SysUtils,
ConTnrs, unitResourceDetails;

type
  TResourceList = class(TResourceModule)
  private
    fResourceList: TObjectList;
  protected
    function GetResourceCount: Integer; override;
    function GetResourceDetails(idx: Integer): TResourceDetails; override;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(src: TResourceModule);
    procedure InsertResource(idx: Integer; details: TResourceDetails); override;
    procedure DeleteResource(idx: Integer); override;
    function AddResource(details: TResourceDetails): Integer; override;
    function IndexOfResource(details: TResourceDetails): Integer; override;
    procedure SortResources; override;
  end;

  TResModule = class(TResourceList)
  private
    f16Bit: boolean;
    procedure ParseResource(header, data: PAnsiChar; dataSize: Integer);
  protected
  public
    procedure SaveToStream(stream: TStream); override;
    procedure LoadFromStream(stream: TStream); override;
  end;

implementation

{ TResModule }

procedure TResModule.ParseResource(header, data: PAnsiChar; dataSize: Integer);
var
  p: PAnsiChar;
  sName, sType: String;
  res: TResourceDetails;
  language, memoryFlags: word;
  version, dataVersion, characteristics: DWORD;

  function GetName: String;
  begin
    if PWord(p)^ = $FFFF then begin
      Inc(p, sizeof(word));
      result := IntToStr(PWord(p)^);
      Inc(p, sizeof(word))
    end else begin
      result := UnicodeString(PWideChar(p));
      Inc(p, (Length(result) + 1) * sizeof(WideChar))
    end
  end;

begin
  try
    p := header;
    Inc(p, 2 * sizeof(Integer));
    sType := GetName;
    sName := GetName;

    if (Integer(p) mod 4) <> 0 then
      Inc(p, 4 - Integer(p) mod 4);

    dataVersion := PDWORD(p)^;
    Inc(p, sizeof(DWORD));
    memoryFlags := PWord(p)^;
    Inc(p, sizeof(word));
    language := PWord(p)^;
    Inc(p, sizeof(word));
    version := PDWORD(p)^;
    Inc(p, sizeof(DWORD));
    characteristics := PDWORD(p)^;
    Inc(p, sizeof(DWORD));

    if (dataSize <> 0) or (sName <> '0') then begin
      res := TResourceDetails.CreateResourceDetails(self, language, sName,
        sType, dataSize, data);
      res.characteristics := characteristics;
      res.version := version;
      res.memoryFlags := memoryFlags;
      res.dataVersion := dataVersion;
      AddResource(res)
    end
    else // NB!!!  32 bit .RES files start with a dummy '32-bit indicator'
      // resource !!!  Is this documented?  I don't think so!

      f16Bit := False;
  except
    raise Exception.Create('The resource file is corrupt');

  end;
end;

procedure TResModule.LoadFromStream(stream: TStream);
var
  buffer, p, q: PAnsiChar;
  bufLen, n, dataSize, HeaderSize, ChunkSize: Integer;
begin
  bufLen := stream.Size;
  GetMem(buffer, bufLen);
  try
    stream.ReadBuffer(buffer^, bufLen); // Read the entite file

    p := buffer;
    n := 0;
    f16Bit := True;
    // Parse each resource
    while n + 2 * sizeof(Integer) < bufLen do begin
      dataSize := PInteger(p)^;
      q := p;
      Inc(q, sizeof(Integer));
      HeaderSize := PInteger(q)^;
      q := p;
      Inc(q, HeaderSize);

      ParseResource(p, q, dataSize);
      ChunkSize := dataSize + HeaderSize;
      ChunkSize := ((ChunkSize + 3) div 4) * 4;
      Inc(p, ChunkSize);
      Inc(n, ChunkSize);
    end;

  finally
    FreeMem(buffer)
  end;
  SortResources
end;

{$POINTERMATH ON}
procedure TResModule.SaveToStream(stream: TStream);
var
  res: TResourceDetails;
  dataSize, HeaderSize, totalSize: Integer;
  header: array [0 .. 1023] of Byte; //char;
  i: Integer;

  function GetResHeader(header: PByte): DWORD;
  var
    pos: DWORD;
    len, dw: DWORD;
    w: word;
    i: Integer;
    ws: WideString;
  begin
    pos := 0;
    ZeroMemory(header, SizeOf(header));

    i := ResourceNameToInt(res.ResourceType);
    if i = -1 then begin
      ws := res.ResourceType;
      len := (Length(ws) + 1) * sizeof(WideChar);
      Move(PWideChar(ws)^, header[pos], len);
      Inc(pos, len)
    end else begin
      w := $FFFF;
      Move(w, header[pos], sizeof(w));
      Inc(pos, sizeof(w));

      w := word(i);
      Move(w, header[pos], sizeof(w));
      Inc(pos, sizeof(w))
    end;

    i := ResourceNameToInt(res.ResourceName);
    if i = -1 then begin
      ws := res.ResourceName;
      len := (Length(ws) + 1) * sizeof(WideChar);
      Move(PWideChar(ws)^, header[pos], len);
      Inc(pos, len)
    end else begin
      w := $FFFF;
      Move(w, header[pos], sizeof(w));
      Inc(pos, sizeof(w));

      w := word(i);
      Move(w, header[pos], sizeof(w));
      Inc(pos, sizeof(w))
    end;

    if (pos mod 4) <> 0 then
      Inc(pos, 4 - (pos mod 4));

    dw := res.dataVersion;
    Move(dw, header[pos], sizeof(DWORD));
    Inc(pos, sizeof(DWORD));

    w := res.memoryFlags;
    Move(w, header[pos], sizeof(word));
    Inc(pos, sizeof(word));

    w := res.ResourceLanguage;
    Move(w, header[pos], sizeof(word));
    Inc(pos, sizeof(word));

    dw := res.version;
    Move(dw, header[pos], sizeof(DWORD));
    Inc(pos, sizeof(DWORD));

    dw := res.characteristics;
    Move(dw, header[pos], sizeof(DWORD));
    Inc(pos, sizeof(DWORD));
    result := pos;
  end;

begin
  if not f16Bit then
  // Write 32-bit resource indicator (An empty type 0 resource)
  begin
    res := TResourceDetails.CreateNew(nil, 0, '0');
    try
      dataSize := res.data.Size;

      stream.WriteBuffer(dataSize, sizeof(dataSize));
      HeaderSize := GetResHeader(PByte(@header[0]));

      totalSize := HeaderSize + 2 * sizeof(DWORD);

      stream.WriteBuffer(totalSize, sizeof(HeaderSize));
      stream.WriteBuffer(header, HeaderSize);
    finally
      res.Free
    end
  end;

  dataSize := 0;
  if ResourceCount > 0 then
    for i := 0 to ResourceCount - 1 do begin
      res := ResourceDetails[i];
      dataSize := res.data.Size;

      stream.WriteBuffer(dataSize, sizeof(dataSize));
      HeaderSize := GetResHeader(PByte(@header[0]));

      totalSize := HeaderSize + 2 * sizeof(DWORD);

      stream.WriteBuffer(totalSize, sizeof(HeaderSize));
      stream.WriteBuffer(header, HeaderSize);
      stream.WriteBuffer(res.data.Memory^, dataSize);

      totalSize := dataSize + totalSize;
      ZeroMemory(@header, sizeof(header));

      if (totalSize mod 4) <> 0 then
        stream.WriteBuffer(header, 4 - (totalSize mod 4));
    end
end;

{ TResourceList }

function TResourceList.AddResource(details: TResourceDetails): Integer;
begin
  result := fResourceList.Add(details);
end;

procedure TResourceList.Assign(src: TResourceModule);
var
  i: Integer;
  res: TResourceDetails;
begin
  fResourceList.Clear;

  for i := 0 to src.ResourceCount - 1 do begin
    res := TResourceDetails.CreateResourceDetails(self,
      src.ResourceDetails[i].ResourceLanguage,
      src.ResourceDetails[i].ResourceName, src.ResourceDetails[i].ResourceType,
      src.ResourceDetails[i].data.Size, src.ResourceDetails[i].data.Memory);

    fResourceList.Add(res)
  end
end;

constructor TResourceList.Create;
begin
  fResourceList := TObjectList.Create;
end;

procedure TResourceList.DeleteResource(idx: Integer);
var
  res: TResourceDetails;
begin
  res := ResourceDetails[idx];
  inherited;
  idx := IndexOfResource(res);
  if idx <> -1 then
    fResourceList.Delete(idx)
end;

destructor TResourceList.Destroy;
begin
  fResourceList.Free;
  inherited;
end;

function TResourceList.GetResourceCount: Integer;
begin
  result := fResourceList.Count
end;

function TResourceList.GetResourceDetails(idx: Integer): TResourceDetails;
begin
  result := TResourceDetails(fResourceList[idx])
end;

function TResourceList.IndexOfResource(details: TResourceDetails): Integer;
begin
  result := fResourceList.IndexOf(details)
end;

procedure TResourceList.InsertResource(idx: Integer; details: TResourceDetails);
begin
  fResourceList.Insert(idx, details)
end;

procedure TResourceList.SortResources;
begin
  fResourceList.Sort(compareDetails);
end;

end.
